/*
 * driver/ide/em86xx/em86xx_bm.c
 * 
 * EM86XX Bus Master IDE driver
 *
 * PIO mode support by Ho Lee 01/15/2004
 * Multi-word DMA mode support by Ho Lee 01/29/2004
 */

/*
 * Tango Bus Master IDE interface is different from that of Japser
 * Japser's IDE controller is compatible wtih PCI-based IDE controller,
 * and Tango's IDE controller is based on Japser's, but it doesn't make 
 * use of PRD table for bus mastering and DMA transfer must go through
 * MBUS. So device driver for Tango bus master IDE controller with DMA
 * support looks similar to that of Mambo's ISA IDE controller. 
 *
 * So this implementation has many duplicated code with em86xx_isa.c
 * But I would like to have separate file to keep clean code for that,
 * because em86xx_isa.c contains some experimental code which is not used,
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <linux/ide.h>
#include <asm/tango2/hardware.h>
#include <asm/tango2/tango2api.h>
#include "em86xx_ide.h"

#ifdef MODULE
#include <linux/module.h>
MODULE_AUTHOR("Craig Qu");
MODULE_DESCRIPTION(" Tango2 BM IDE driver");
MODULE_LICENSE("GPL");
#endif


#ifdef CONFIG_BLK_DEV_TANGO2_BMIDEDMA
// handlers
static int em86xx_bmide_dma_read(ide_drive_t *drive);
static int em86xx_bmide_dma_write(ide_drive_t *drive);
static int em86xx_bmide_dma_begin(ide_drive_t *drive);
static int em86xx_bmide_dma_end(ide_drive_t *drive);
static int em86xx_bmide_dma_check(ide_drive_t *drive);
static int em86xx_bmide_dma_host_on(ide_drive_t *drive);
static int em86xx_bmide_dma_host_off(ide_drive_t *drive);
static int em86xx_bmide_tune_chipset(ide_drive_t *drive, u8 speed);
static int em86xx_bmide_dma_lostirq(ide_drive_t *drive);
static int em86xx_bmide_dma_timeout(ide_drive_t *drive);

// helper functions
static struct buffer_head *em86xx_bmide_dma_setup(int read, struct request *rq, int stop_if_fragment);
static int em86xx_bmide_rwdma(ide_drive_t *drive, int read, struct request *rq);
static int em86xx_bmide_dma_end_io(ide_drive_t *drive);
static ide_startstop_t em86xx_bmide_dma_intr(ide_drive_t *drive);

// switchbox and MBUS
static int g_sbox_port_read = -1, g_sbox_port_write = -1;
static unsigned int g_regbase_read, g_regbase_write;
static int g_irq_read, g_irq_write;

// Misc
static struct request *crq = NULL;
static int crq_ntransfer = 0;

/* Merge BH as much as possible */
#define EM86XX_BMIDE_DMA_MERGE_BH

// bounce buffer
#define DMA_BOUNCE_BUF_ORDER		(19 - PAGE_SHIFT)	// maximum = 128 sectors = 2^7 sectors = 2^7 * 2^9 bytes = 2^16 bytes
static unsigned char *g_bounce_buf = NULL;
#endif

// helper functions
static void em86xx_bmide_tune_chipset_drive(int drive, int speed, int verbose);

//
// data structure initialization
//

void em86xx_bmide_init_hwif_ports(hw_regs_t *hw, ide_ioreg_t data_port, ide_ioreg_t ctrl_port, int *irq)
{
	int i;
	ide_ioreg_t reg;

	memset(hw, 0, sizeof(*hw));

	// setup io_ports
	for (i = IDE_DATA_OFFSET, reg = data_port; i <= IDE_STATUS_OFFSET; i++) {
		hw->io_ports[i] = reg;
		reg += 4;
	}

	hw->io_ports[IDE_CONTROL_OFFSET] = ctrl_port;
	hw->dma = NO_DMA;

	// Tango has two interrupts for IDE controller : IRQ_IDECTRL_IDE, IRQ_IDECTRL_IDEDMA
	// IRQ_IDECTRL_IDEDMA is to be used for DMA transfer but it is almost same with 
	// IRQ_IDECTRL_IDE interrupt.  Tango triggers IDEDMA interrupt after DMA FIFO is
	// flushed, and it stays high until BMIS register bit 2 is cleared. 
	// Use IRQ_IDECTRL_IDE interrupt for both of PIO and DMA.
	if (irq) 
		*irq = IRQ_IDECTRL_IDE;
}

//
// initialize IDE driver
//

int __init em86xx_bmide_init(void)
{
	int index, i;
	ide_hwif_t *hwif;
	hw_regs_t hw;

#if defined(CONFIG_TANGO2_SIG_BLOCK) || defined(CONFIG_TANGO2_XENV)
	extern int tango2_bmide_enabled(void);
	if (tango2_bmide_enabled() == 0)
		return 0;
#endif
	// Enable bus master IDE interface
	gbus_write_uint32(pGBus, REG_BASE_host_interface + IDECTRL_pri_idectl, 0x8400);
	// no skew : normal IORDY, normal data
	gbus_write_uint32(pGBus, REG_BASE_host_interface + IDECTRL_idesrc, 0);
	// Enable non-PRD mode
	gbus_write_uint32(pGBus, REG_BASE_host_interface + IDECTRL_bmic, 0x4);

	em86xx_bmide_init_hwif_ports(&hw, REG_BASE_host_interface_BMIDE, REG_BASE_host_interface + IDECTRL_pri_ctrlblock, &hw.irq);
	index = ide_register_hw(&hw, NULL);
	printk("SMP863x Bus Mastering IDE activated as ide%d.\n", index);

	// initialize ide_hwifs data structure
	// supports only one interface
	for (i = index; i < index + 1; ++i) {
		// register DMA handler
		hwif = &ide_hwifs[i];
		hwif->OUTB = em86xx_ide_OUTB;
		hwif->OUTW = em86xx_ide_OUTW;
		hwif->OUTBSYNC = em86xx_ide_OUTBSYNC;
		hwif->INB = em86xx_ide_INB;
		hwif->INW = em86xx_ide_INW;
		hwif->INSW = em86xx_ide_INSW;
		hwif->OUTSW = em86xx_ide_OUTSW;
                // MBUS can transfer two separate chunks of data blocks
                // but kernel usually allocate 4KB for each buffer cache,
                // which can hold 8 sectors, and usually they are not continous
                // So TANGO2X IDE interface can handle 16 sectors in best performance
                // If the request queue is bigger than this, driver need to use
                // bounce buffer
		hwif->rqsize = 16;
	    if (!noautodma) {
#ifdef CONFIG_BLK_DEV_TANGO2_BMIDEDMA
		hwif->autodma = 1;
		hwif->atapi_dma = 1;
#ifdef CONFIG_BLK_DEV_TANGO2_BMIDEUDMA
		hwif->ultra_mask = 0x7f;        // enable all Ultra DMA	
#else
		hwif->ultra_mask = 0x00;		// disable all Ultra DMA
#endif
		hwif->udma_four = 1;
		hwif->mwdma_mask = 0x07;		// enable all Multi-word DMA (0/1/2)
		hwif->swdma_mask = 0x80;		// disable all Single-word DMA

		// registers
		hwif->dma_command = REG_BASE_host_interface + IDECTRL_bmic;
		hwif->dma_status = REG_BASE_host_interface + IDECTRL_bmis;
		hwif->dma_prdtable = REG_BASE_host_interface + IDECTRL_bmidtp;

		// handlers
		hwif->ide_dma_read = em86xx_bmide_dma_read;
		hwif->ide_dma_write = em86xx_bmide_dma_write;
		hwif->ide_dma_begin = em86xx_bmide_dma_begin;
		hwif->ide_dma_end = em86xx_bmide_dma_end;
		hwif->ide_dma_check = em86xx_bmide_dma_check;
		hwif->ide_dma_on = em86xx_ide_dma_on;
		hwif->ide_dma_off = em86xx_ide_dma_off;
		hwif->ide_dma_off_quietly = em86xx_ide_dma_off_quietly;
		hwif->ide_dma_test_irq = em86xx_ide_dma_test_irq;
		hwif->ide_dma_host_on = em86xx_bmide_dma_host_on;
		hwif->ide_dma_host_off = em86xx_bmide_dma_host_off;
		hwif->ide_dma_bad_drive = em86xx_ide_dma_bad_drive;
		hwif->ide_dma_good_drive = em86xx_ide_dma_good_drive;
		hwif->ide_dma_count = em86xx_ide_dma_count;
		hwif->ide_dma_verbose = em86xx_ide_dma_verbose;
		hwif->ide_dma_retune = em86xx_ide_dma_retune;
		hwif->ide_dma_lostirq = em86xx_bmide_dma_lostirq;
		hwif->ide_dma_timeout = em86xx_bmide_dma_timeout;
		hwif->speedproc = em86xx_bmide_tune_chipset;
#endif
	    }
	}

	// set to maximum PIO mode
	for (i = 0; i < MAX_DRIVES; ++i)
		em86xx_bmide_tune_chipset_drive(i, XFER_PIO_4, 0);

	if (!noautodma) {
#ifdef CONFIG_BLK_DEV_TANGO2_BMIDEDMA
	if ((g_bounce_buf = (unsigned char *)__get_free_pages(GFP_KERNEL, DMA_BOUNCE_BUF_ORDER)) == NULL)
		printk("IDE: Can not allocate buffer for IDE transfer\n");

#ifdef EM86XX_BOUNCE_BUFFER_NOCACHE
	g_bounce_buf = (unsigned char *)KSEG1ADDR((unsigned long)g_bounce_buf);
	printk("Bounce buffer starts at 0x%08lx\n", (unsigned long)g_bounce_buf);
#endif

#endif	// EM86XX_BMIDEDMA
	}
	return 0;
}

#ifdef CONFIG_BLK_DEV_TANGO2_BMIDEDMA

//
// IDE DMA handlers
//

int em86xx_bmide_dma_read(ide_drive_t *drive)
{
	struct request *rq;

	drive->waiting_for_dma = 1;
	rq = HWGROUP(drive)->rq;
	// printk("em86xx_bmide_dma_read : nsect = %ld\n", rq->nr_sectors);
	return em86xx_bmide_rwdma(drive, 1, rq);
}

int em86xx_bmide_dma_write(ide_drive_t *drive)
{
	struct request *rq;

	drive->waiting_for_dma = 1;
	rq = HWGROUP(drive)->rq;
	// printk("em86xx_bmide_dma_write : sector = %ld, nsect = %ld\n", rq->sector, rq->nr_sectors);
	return em86xx_bmide_rwdma(drive, 0, rq);
}

int em86xx_bmide_dma_begin(ide_drive_t *drive)
{
	return(0);
}

int em86xx_bmide_dma_end(ide_drive_t *drive)
{
	int status;

	drive->waiting_for_dma = 0;
	status = em86xx_bmide_dma_end_io(drive);
	return(status);
}

int em86xx_bmide_dma_lostirq(ide_drive_t *drive)
{
	struct request *rq = HWGROUP(drive)->rq;
	unsigned int regbase = ((rq->cmd == READ) ? g_regbase_read : g_regbase_write);

	if (em86xx_mbus_inuse(regbase) != 0) 
		em86xx_mbus_reset(regbase, SBOX_IDEDVD);
	return 0;
}

int em86xx_bmide_dma_timeout(ide_drive_t *drive)
{
	struct request *rq = HWGROUP(drive)->rq;
	unsigned int regbase = ((rq->cmd == READ) ? g_regbase_read : g_regbase_write);

	if (em86xx_mbus_inuse(regbase) != 0) 
		em86xx_mbus_reset(regbase, SBOX_IDEDVD);
	return 0;
}

int em86xx_bmide_dma_check(ide_drive_t *drive)
{
	int mode;

#ifdef CONFIG_BLK_DEV_TANGO2_BMIDEUDMA
	if ((gbus_read_uint32(pGBus, REG_BASE_host_interface + IDECTRL_idestatus) & 0x02) == 0x02)
        	mode = 1;   // 40 donductor cable (UDMA33) : Mode 0/1/2
	else 
		mode = 4;   // 80 conductor cable : Mode 0/1/2/3/4/5/6
#else
	mode = 0;       // MW DMA mode 2
#endif

	return em86xx_ide_config_drive_for_dma(drive, mode);
}

int em86xx_bmide_dma_host_on(ide_drive_t *drive)
{
	if (drive->using_dma) {
		ide_hwif_t *hwif = HWIF(drive);
		u8 unit = (drive->select.b.unit & 0x01);
		u8 dma_stat = hwif->INB(hwif->dma_status);

		hwif->OUTB((dma_stat | (1 << (5 + unit))), hwif->dma_status);
		return 0;
	}
	return 1;
}

int em86xx_bmide_dma_host_off(ide_drive_t *drive)
{
	ide_hwif_t *hwif = HWIF(drive);
	u8 unit = (drive->select.b.unit & 0x01);
	u8 dma_stat = hwif->INB(hwif->dma_status);

	hwif->OUTB((dma_stat & ~(1 << (5 + unit))), hwif->dma_status);

	return 0;
}

int em86xx_bmide_tune_chipset(ide_drive_t *drive, u8 speed)
{
	em86xx_bmide_tune_chipset_drive(drive->select.b.unit, speed, 1);

	return ide_config_drive_speed(drive, speed);
}

//
// helper functions
//

#ifdef EM86XX_BMIDE_DMA_MERGE_BH
#define MBUS_LINEAR_MAX 	(0x2000 - 1)
#endif

/* Get the I/O size of given request */
static inline int get_request_size(struct request *rq)
{
	unsigned int rq_size;
	struct buffer_head *bh;

	for (rq_size = 0, bh = rq->bh; bh != NULL; bh = bh->b_reqnext) 
		rq_size += bh->b_size;

	return(rq_size);
}

#ifdef CONFIG_BLK_DEV_TANGO2_BMIDEDMA
// return the number of fragmented blocks
static __inline__ int em86xx_bmide_request_fragment(struct request *rq)
{
	int ntransfer = 0;
	struct buffer_head *bh = rq->bh;

#ifdef EM86XX_BMIDE_DMA_MERGE_BH
	unsigned long addr;
	unsigned int size;
	unsigned int sector;

	if (rq == crq)
		return(crq_ntransfer);

	/* Try to merge multiple BH if possible -- deal with one fragment only */
	for (ntransfer = 0, bh = rq->bh;;) {
		if (bh == NULL)
			break;
		else {
			if (++ntransfer > 1) /* More than one fragment found */
				break; 
		}
		addr = (unsigned long)bh->b_data;
		size = bh->b_size;
		sector = bh->b_rsector;
		for (bh = bh->b_reqnext; bh != NULL; bh = bh->b_reqnext) {
			if ((addr + size) != (unsigned long)bh->b_data)
				break;	/* data is not contiguous */
			else {	/* Merge the next BH */
				size += bh->b_size;
			}
		}	
	}
	if (ntransfer < 2) /* This can be done with one fragment (linear or rectangular) */
		goto done;

	/* Try to merge multiple BH if possible -- deal with two fragments or more */
	for (ntransfer = 0, bh = rq->bh;;) {
		if (bh == NULL)
			break;
		else
			++ntransfer;
		addr = (unsigned long)bh->b_data;
		size = bh->b_size;
		sector = bh->b_rsector;
		for (bh = bh->b_reqnext; bh != NULL; bh = bh->b_reqnext) {
			if ((addr + size) != (unsigned long)bh->b_data)
				break;	/* data is not contiguous */
			else if ((size + bh->b_size) > MBUS_LINEAR_MAX)
				break;	/* Size too large */
			else {	/* Merge the next BH */
				size += bh->b_size;
			}
		}	
	}
#else
	if (rq == crq)
		return(crq_ntransfer);

	do {
		++ntransfer;
	} while ((bh = bh->b_reqnext) != NULL);
#endif

done:
	return ntransfer;
}

// prepare MBUS for DMA transfer
// stop_if_fragment must be non-zero
static struct buffer_head *em86xx_bmide_dma_setup(int read, struct request *rq, int stop_if_fragment)
{
	struct buffer_head *bh = rq->bh;

	// scans buffer head and calculate the requested transfer size
	// MBUS can't transfer to more than two separate regions. 
	// So if the buffer head is too fragmented, return non-NULL value
#ifdef EM86XX_BMIDE_DMA_MERGE_BH
	struct {
		unsigned long addr;
		unsigned int size;
		unsigned int sector;
	} transfer[2], *ptransfer;
	unsigned int ntransfer;

	crq = rq;

	/* Try to merge multiple BH if possible -- one fragment only */
	for (ntransfer = 0, ptransfer = &transfer[0], bh = rq->bh;; ptransfer++) {
		if (bh == NULL)
			break;
		else if (++ntransfer > 1)
			break;

		ptransfer->addr = (unsigned long)bh->b_data;
		ptransfer->size = bh->b_size;
		ptransfer->sector = bh->b_rsector;
		for (bh = bh->b_reqnext; bh != NULL; bh = bh->b_reqnext) {
			if ((ptransfer->addr + ptransfer->size) != (unsigned long)bh->b_data)
				break;	/* data is not contiguous */
			else {	/* Merge the next BH */
				ptransfer->size += bh->b_size;
			}
		}	
	}
	if (ntransfer < 2) /* This can be done with one fragment (linear or rectangular) */
		goto start_transfer;

	/* Try to merge multiple BH if possible */
	for (ntransfer = 0, ptransfer = &transfer[0], bh = rq->bh;; ptransfer++) {
		if (bh == NULL)
			break;
		else if (++ntransfer > 2)
			break;

		ptransfer->addr = (unsigned long)bh->b_data;
		ptransfer->size = bh->b_size;
		ptransfer->sector = bh->b_rsector;
		for (bh = bh->b_reqnext; bh != NULL; bh = bh->b_reqnext) {
			if ((ptransfer->addr + ptransfer->size) != (unsigned long)bh->b_data)
				break;	/* data is not contiguous */
			else if ((ptransfer->size + bh->b_size) > MBUS_LINEAR_MAX)
				break;	/* Size too large */
			else {	/* Merge the next BH */
				ptransfer->size += bh->b_size;
			}
		}	
	}

	if ((ntransfer > 2) && (stop_if_fragment != 0)) {
		crq_ntransfer = ntransfer;
		return bh;
	}
#else
	struct buffer_head *prevbh;

	struct {
		unsigned long addr;
		unsigned int size;
	} transfer[2], *ptransfer;
	unsigned int ntransfer, size;

	crq = rq;

	for (prevbh = NULL, ntransfer = 0, ptransfer = transfer, size = 0; bh != NULL; prevbh = bh, bh = bh->b_reqnext) {
		if (++ntransfer > 2) {
			if (stop_if_fragment) {
				crq_ntransfer = ntransfer;
				return bh;
			}
			break;
		}
		ptransfer->addr = (unsigned long) bh->b_data;
		ptransfer->size = bh->b_size;
		// printk("MBUS : bh = %p, bh->next = %p, %08x, %08x\n", bh, bh->b_reqnext, ptransfer->addr, ptransfer->size);
		++ptransfer;
		size += bh->b_size;
	}
#endif

start_transfer:

	crq_ntransfer = ntransfer;

	// setup MBUS
	if (ntransfer == 1) {
#ifdef CONFIG_NONCOHERENT_IO
		if (!is_physical(transfer[0].addr)) {
			if (read) 
				dma_cache_inv(transfer[0].addr, transfer[0].size);
			else
				dma_cache_wback_inv(transfer[0].addr, transfer[0].size);
		}
#endif
		if (transfer[0].size <= MBUS_LINEAR_MAX) {
			/* Do linear transfer */
			em86xx_mbus_setup_dma_linear(read ? g_regbase_read : g_regbase_write, SBOX_IDEDVD,
					is_physical(transfer[0].addr) ? transfer[0].addr : PHYSADDR(transfer[0].addr), 
					transfer[0].size);
		} else {
			/* Do rectangular transfer */
			unsigned int i, horiz, lines, sz;
			for (i = 0, horiz = 1, sz = transfer[0].size; (i < 12) && ((sz & 0x01) == 0); ++i, horiz <<= 1, sz >>= 1) 
				;
			lines = transfer[0].size >> i;
			em86xx_mbus_setup_dma_rectangle(read ? g_regbase_read : g_regbase_write, SBOX_IDEDVD,
					is_physical(transfer[0].addr) ? transfer[0].addr : PHYSADDR(transfer[0].addr), 
					horiz, lines, horiz);
		}
	} else {
#ifdef CONFIG_NONCOHERENT_IO
		if (read) {
			if (!is_physical(transfer[0].addr)) 
				dma_cache_inv(transfer[0].addr, transfer[0].size);
			if (!is_physical(transfer[1].addr)) 
				dma_cache_inv(transfer[1].addr, transfer[1].size);
		} else {
			if (!is_physical(transfer[0].addr)) 
				dma_cache_wback_inv(transfer[0].addr, transfer[0].size);
			if (!is_physical(transfer[1].addr)) 
				dma_cache_wback_inv(transfer[1].addr, transfer[1].size);
		}
#endif
		em86xx_mbus_setup_dma_double(read ? g_regbase_read : g_regbase_write, SBOX_IDEDVD,
			is_physical(transfer[0].addr) ? transfer[0].addr : PHYSADDR(transfer[0].addr), transfer[0].size,
			is_physical(transfer[1].addr) ? transfer[1].addr : PHYSADDR(transfer[1].addr), transfer[1].size);
	}

	return bh;
}

// core of DMA read and write transfer
int em86xx_bmide_rwdma(ide_drive_t *drive, int read, struct request *rq)
{
	unsigned int buflen = get_request_size(rq);
	unsigned long cj;
	int retry;
	static const int max_retry = 5;

	// setup switchbox and wait for its readiness
	if (g_sbox_port_read < 0) {
		for (retry = 0; ((g_sbox_port_read = em86xx_mbus_alloc_dma(SBOX_IDEDVD, 1, &g_regbase_read, &g_irq_read)) < 0) && (retry < max_retry); retry++) {
			for (cj = jiffies + (HZ / 100); jiffies < cj;);
		}
		if (g_sbox_port_read < 0) { /* Failed to use DMA, use PIO instead */
			printk("%s: failed to obtain DMA channel, use PIO instead.\n", drive->name);
			return(1);
		}
	}
	if (g_sbox_port_write < 0) {
		for (retry = 0; ((g_sbox_port_write = em86xx_mbus_alloc_dma(SBOX_IDEDVD, 0, &g_regbase_write, &g_irq_write)) < 0) && (retry < max_retry); retry++) {
			for (cj = jiffies + (HZ / 100); jiffies < cj;);
		}
		if (g_sbox_port_write < 0) { /* Failed to use DMA, use PIO instead */
			printk("%s: failed to obtain DMA channel, use PIO instead.\n", drive->name);
			em86xx_mbus_free_dma(g_sbox_port_read);
			g_sbox_port_read = -1;
			return(1);
		}
	}

	// setup interrupt handler
	// only required for hard disk. CD-ROM driver registers a handler for itself
	if (drive->media == ide_disk)
		ide_set_handler(drive, &em86xx_bmide_dma_intr, 10 * WAIT_CMD, NULL);

	// setup MBUS
	if (em86xx_bmide_dma_setup(read, rq, 1)) {
		// if the request is too fragmented (over two non-continuous regions)
		// use the bounce buffer. For read transfers, transfer to bounce buffer
		// and copy to original buffer head after transfer. For write transfers,
		// copy the data to be written to bounce buffer first.
		unsigned long buf = (unsigned long) g_bounce_buf;

		if (!read) {
			struct buffer_head *bh = rq->bh;
#if 1
			char *data;
			u32 size;
			extern int mbus_memcpy(u32 dst, u32 src, u32 size);
			
			/* Try to merge multiple BH for copying */
			while (bh != NULL) {
				data = bh->b_data;
				size = bh->b_size;
				for (bh = bh->b_reqnext; bh != NULL; bh = bh->b_reqnext) {
					if ((data + size) == bh->b_data)
						size += bh->b_size; /* Contiguous */
					else 
						break;
				}
				mbus_memcpy(PHYSADDR(buf), PHYSADDR(data), size);
#ifdef CONFIG_NONCOHERENT_IO
				dma_cache_inv(buf, size);
#endif
				buf += size;
			}
#else
			do {
				memcpy((void *) buf, bh->b_data, bh->b_size);
				buf += bh->b_size;
			} while ((bh = bh->b_reqnext) != NULL);
#endif
#ifdef CONFIG_SD_CHECKERROR
			if (buf > ((unsigned long) g_bounce_buf + (1 << (DMA_BOUNCE_BUF_ORDER + PAGE_SHIFT))))
				printk("IDE: Bounce buffer is too small to fit requested transfers\n");
#endif
		}

#ifdef CONFIG_NONCOHERENT_IO
#ifndef EM86XX_BOUNCE_BUFFER_NOCACHE
		dma_cache_wback_inv((unsigned long)g_bounce_buf, buflen);
#endif
#endif
		// try to setup MBUS again. This always succeeds
		em86xx_mbus_setup_dma(read ? g_regbase_read : g_regbase_write, SBOX_IDEDVD, PHYSADDR((unsigned long)g_bounce_buf), buflen, NULL, NULL);
	} else {
		em86xx_ide_dma_cache(rq, read);
	}

	// setup IDE DMA pointer and length
	// IDE DMA pointer is not important since IDE controller is just a tunnel
	// to MBUS. 
	// gbus_write_uint32(pGBus, REG_BASE_host_interface + IDECTRL_ide_dmaptr, (unsigned long)buf);
	gbus_write_uint32(pGBus, REG_BASE_host_interface + IDECTRL_ide_dmalen, (unsigned long)buflen);

	// send IDE command
	// 28 bits : READDMA or WRITEDMA
	// 48 bits : READDMA_EXT or WRITEDMA_EXT
	// DMARQ is asserted to low
	if (drive->media == ide_disk) {
		HWIF(drive)->OUTB(read ? (drive->addressing ? WIN_READDMA_EXT : WIN_READDMA) : (drive->addressing ? WIN_WRITEDMA_EXT : WIN_WRITEDMA), IDE_COMMAND_REG);
	}

	// start bus master transfer
	// non-prd mode & bus master start
	// DMAACK is asserted to high
	gbus_write_uint32(pGBus, REG_BASE_host_interface + IDECTRL_bmic, 0x05 | (read ? 0x08 : 0x00));

	return 0;
}
#endif

// cleanup after DMA transfer is done
int em86xx_bmide_dma_end_io(ide_drive_t *drive)
{
	int stat = 0;
	struct request *rq = HWGROUP(drive)->rq;

	// wait for DMA interrupt
	// em86xx_wait_irq(IRQ_IDECTRL_IDEDMA);
	// clear DMA interrupt

	for (stat = 0; stat < 2000000; stat++) {
		if ((gbus_read_uint32(pGBus, REG_BASE_host_interface + IDECTRL_bmis) & 0x4) != 0)
			break;
		else
			udelay(1);
	}

	if (stat >= 2000000)
		printk("%s: dma timeout.\n", drive->name);

	gbus_write_uint32(pGBus, REG_BASE_host_interface + IDECTRL_bmis, gbus_read_uint32(pGBus, REG_BASE_host_interface + IDECTRL_bmis));

	// confirm whether MBUS transfer is done
	// due to the memory arbitration, IDE device thinks the DMA transfer
	// is done, but the data might be held in MBUS FIFO
	stat = em86xx_mbus_wait(rq->cmd == READ ? g_regbase_read : g_regbase_write, SBOX_IDEDVD);

	HWIF(drive)->INB(IDE_STATUS_REG);

	// Stop bus mastering
	gbus_write_uint32(pGBus, REG_BASE_host_interface + IDECTRL_bmic, 0x4);

	// Check status
	if (stat != 0) 
		goto done;
	
	if (rq->cmd == READ) {
		if (em86xx_bmide_request_fragment(rq) > 2) {
			// copy buffer after transfer
			unsigned long buf = (unsigned long) g_bounce_buf;
			struct buffer_head *bh = rq->bh;
#ifdef CONFIG_NONCOHERENT_IO
#ifndef EM86XX_BOUNCE_BUFFER_NOCACHE
			dma_cache_inv(buf, get_request_size(rq));
#endif
#endif
#if 1
			{
				char *data;
				u32 size;
				extern int mbus_memcpy(u32 dst, u32 src, u32 size);
			
				/* Try to merge multiple BH for copying */
				while (bh != NULL) {
					data = bh->b_data;
					size = bh->b_size;
					for (bh = bh->b_reqnext; bh != NULL; bh = bh->b_reqnext) {
						if ((data + size) == bh->b_data)
							size += bh->b_size; /* Contiguous */
						else 
							break;
					}
					mbus_memcpy(PHYSADDR(data), PHYSADDR(buf), size);
#ifdef CONFIG_NONCOHERENT_IO
					if (!is_physical((u32)data))
						dma_cache_inv((unsigned long)data, size);
#endif
					buf += size;
				}
			}
#else
			do {
				memcpy(bh->b_data, (void *) buf, bh->b_size);
				buf += bh->b_size;
			} while ((bh = bh->b_reqnext) != NULL);
#endif
		} else {
			em86xx_ide_dma_cache(rq, 1 /* READ */);
		}
	}

done:
	/* Free up MBUS channel */
	em86xx_mbus_free_dma(g_sbox_port_read);
	em86xx_mbus_free_dma(g_sbox_port_write);
	g_sbox_port_read = g_sbox_port_write = -1;
	return((stat != 0) ? 1 : 0);
}

// DMA interrupt handler after transfer is done
ide_startstop_t em86xx_bmide_dma_intr(ide_drive_t *drive)
{
	int i;
	ide_hwif_t *hwif = HWIF(drive);
	struct request *rq = HWGROUP(drive)->rq;

	hwif->ide_dma_end(drive);

	for (i = rq->nr_sectors; i > 0;) {
		i -= rq->current_nr_sectors;
		ide_end_request(drive, 1);
	}

	return ide_stopped;
}

#endif

void em86xx_bmide_tune_chipset_drive(int drive, int speed, int verbose)
{
	int mode = speed & 0x0f;
	unsigned int ctrlreg = gbus_read_uint32(pGBus, REG_BASE_host_interface + IDECTRL_pri_idectl);
	unsigned int newflag = 0;
	
	if (speed >= XFER_PIO_0 && speed <= XFER_PIO_4) {
		// setup timing for PIO mode
		static unsigned int s_pio_tim[] = {
			0xff230ee6, 0xd41b0fa4, 0xb4150f63, 0xa7110f62, 0x9a0f0552
		};

		mode -= 8;
		if (verbose)
			printk("IDE: Set drive %d to PIO mode %d\n",
				drive, mode);

		// fast timing for PIO
		newflag = 0x01;
		
		gbus_write_uint32(pGBus, REG_BASE_host_interface + ((drive == 0) ? IDECTRL_pri_drv0tim : IDECTRL_pri_drv1tim), s_pio_tim[mode]);
	} 
#ifdef CONFIG_BLK_DEV_TANGO2_BMIDEDMA
	else if (speed >= XFER_MW_DMA_0 && speed <= XFER_MW_DMA_2) {
		// setup timing for Multi-word DMA
		static unsigned int s_dma_tim[] = {
			0xe92f2fa4, 0xa0110a63, 0x9a0f0552
		};
		static unsigned int s_udma_tim1[] = {
			0x33330f0f, 0x33330f0f, 0x33330f0f,
		};
		static unsigned int s_udma_tim2[] = {
			0x00000107, 0x00000107, 0x00000107,
		};

		if (verbose)
			printk("IDE: Set drive %d to Multi-word DMA mode %d\n",
				drive, mode);

		// fast timing for PIO, prefetch enable
		newflag = 0x05;

		gbus_write_uint32(pGBus, REG_BASE_host_interface + ((drive == 0) ? IDECTRL_pri_drv0tim : IDECTRL_pri_drv1tim), s_dma_tim[mode]);
		gbus_write_uint32(pGBus, REG_BASE_host_interface + ((drive == 0) ? IDECTRL_pri_drv0udmatim1 : IDECTRL_pri_drv1udmatim1), s_udma_tim1[mode]);
		gbus_write_uint32(pGBus, REG_BASE_host_interface + ((drive == 0) ? IDECTRL_pri_drv0udmatim2 : IDECTRL_pri_drv1udmatim2), s_udma_tim2[mode]);
	} 
#endif
#ifdef CONFIG_BLK_DEV_TANGO2_BMIDEUDMA
	else if (speed >= XFER_UDMA_0 && speed <= XFER_UDMA_7) {
		// setup timing for Ultra DMA
		static unsigned int s_udma_tim1[] = {
			0x44442418, 0x44441c10, 0x4444160b, 0x44441608,
			0x44441605, 0x44441303, 0x44441302, 0x44441302
		};
		static unsigned int s_udma_tim2[] = {
			0x0000110f, 0x0000110a, 0x00001106, 0x00001104, 
			0x00001101, 0x00001202, 0x00001202, 0x00001202
		};

		if (verbose)
			printk("IDE: Set drive %d to Ultra DMA mode %d\n",
				drive, mode);

		newflag = 0x00;

		// enable Ultra DMA
		gbus_write_uint32(pGBus, REG_BASE_host_interface + IDECTRL_udmactl, 
			gbus_read_uint32(pGBus, REG_BASE_host_interface + IDECTRL_udmactl) | ((drive == 0) ? 0x01 : 0x02));
		
		// set timing register
		gbus_write_uint32(pGBus, REG_BASE_host_interface + ((drive == 0) ? IDECTRL_pri_drv0udmatim1 : IDECTRL_pri_drv1udmatim1), s_udma_tim1[mode]);
		gbus_write_uint32(pGBus, REG_BASE_host_interface + ((drive == 0) ? IDECTRL_pri_drv0udmatim2 : IDECTRL_pri_drv1udmatim2), s_udma_tim2[mode]);
	}
#endif

	ctrlreg &= ~(drive == 0 ? 0x0f : 0xf0);
	ctrlreg |= newflag << (drive * 4);
	gbus_write_uint32(pGBus, REG_BASE_host_interface + IDECTRL_pri_idectl, ctrlreg);
}



void __exit em86xx_bmide_exit(void)
{

#ifdef CONFIG_BLK_DEV_TANGO2_BMIDEDMA
#ifdef EM86XX_BOUNCE_BUFFER_NOCACHE
        g_bounce_buf = (unsigned char *)em86xx_to_caddr((unsigned long)g_bounce_buf);
#endif
        free_pages((unsigned long)g_bounce_buf, DMA_BOUNCE_BUF_ORDER);
#endif
        return;
}


#ifdef MODULE
module_init(em86xx_bmide_init);
module_exit(em86xx_bmide_exit);
#endif

