#include <linux/config.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/ioport.h>
#include <linux/spinlock.h>
#include <linux/pm.h>
#include <linux/kmod.h>		/* for hotplug_path */
#include <linux/bitops.h>
#include <linux/delay.h>
#include <linux/cache.h>
#include <linux/pci.h>

#include <asm/page.h>
#include <asm/dma.h>	/* isa_dma_bridge_buggy */

#ifdef CONFIG_TANGO2
#include <asm/tango2/hardware.h>
#include "tango2-usb.h"
#endif

#undef DEBUG

#ifdef DEBUG
#define DBG(x...) printk(x)
#else
#define DBG(x...)
#endif

/*
 * Pool allocator ... wraps the tango2_alloc_consistent page allocator, so
 * small blocks are easily used by drivers for bus mastering controllers.
 * This should probably be sharing the guts of the slab allocator.
 */

struct tango2_pool {    /* the pool */
        struct list_head        page_list;
        spinlock_t              lock;
        size_t                  blocks_per_page;
        size_t                  size;
        int                     flags;
        //struct pci_dev                *dev;
        size_t                  allocation;
        char                    name [32];
        wait_queue_head_t       waitq;
};

struct tango2_page {    /* cacheable header for 'allocation' bytes */
        struct list_head        page_list;
        void                    *vaddr;
        dma_addr_t              dma;
        unsigned long           bitmap [0];
};



/*
 * Map a single buffer of the indicated size for DMA in streaming mode.
 * The 32-bit bus address to use is returned.
 *
 * Once the device is given the dma address, the device owns this memory
 * until either pci_unmap_single or pci_dma_sync_single is performed.
 */
dma_addr_t tango2_map_single(/*struct pci_dev *hwdev, */void *ptr,
                                        size_t size, int direction)
{
        unsigned long addr = (unsigned long) ptr;

        if (direction == PCI_DMA_NONE)
                out_of_line_bug();

        dma_cache_wback_inv(addr, size);
	addr = PHYSADDR(ptr);
        return (dma_addr_t)addr;
}

/*
 * Unmap a single streaming mode DMA translation.  The dma_addr and size
 * must match what was provided for in a previous pci_map_single call.  All
 * other usages are undefined.
 *
 * After this call, reads by the cpu to the buffer are guarenteed to see
 * whatever the device wrote there.
 */
void tango2_unmap_single(/*struct pci_dev *hwdev,*/ dma_addr_t dma_addr,
                                    size_t size, int direction)
{

        if (direction == PCI_DMA_NONE)
                out_of_line_bug();

        if (direction != PCI_DMA_TODEVICE) {
                unsigned long addr;
                addr = (unsigned long)phys_to_virt(dma_addr);
                dma_cache_wback_inv(addr, size);
        }
}





void *tango2_alloc_consistent(size_t size, dma_addr_t * dma_handle)
{
        void *ret;
        int gfp = GFP_ATOMIC | GFP_DMA;

        ret = (void *) __get_free_pages(gfp , get_order(size));
//	printk("tango2 alloc_consistent ret=0x%x gfp=0x%x size=0x%x\n", ret, gfp, size);
        if (ret != NULL) {
                memset(ret, 0, size);
                *dma_handle = PHYSADDR((u32)ret);
#ifdef CONFIG_NONCOHERENT_IO
                dma_cache_wback_inv((unsigned long) ret, size);
                ret = UNCAC_ADDR(ret);
#endif
        }
        return ret;
}

void tango2_free_consistent(size_t size, void *vaddr, dma_addr_t dma_handle)
{
        unsigned long addr = (unsigned long) vaddr;

#ifdef CONFIG_NONCOHERENT_IO
        addr = CAC_ADDR(addr);
#endif
        free_pages(addr, get_order(size));
}




#define	POOL_TIMEOUT_JIFFIES	((100 /* msec */ * HZ) / 1000)
#define	POOL_POISON_BYTE	0xa7

//#define CONFIG_TANGO2POOL_DEBUG


/**
 * tango2_pool_create - Creates a pool of consistent memory blocks, for dma.
 * @name: name of pool, for diagnostics
 * @pdev:  device that will be doing the DMA
 * @size: size of the blocks in this pool.
 * @align: alignment requirement for blocks; must be a power of two
 * @allocation: returned blocks won't cross this boundary (or zero)
 * @flags: SLAB_* flags (not all are supported).
 *
 * Returns a  allocation pool with the requested characteristics, or
 * null if one can't be created.  Given one of these pools, tango2_pool_alloc()
 * may be used to allocate memory.  Such memory will all have "consistent"
 * DMA mappings, accessible by the device and its driver without using
 * cache flushing primitives.  The actual size of blocks allocated may be
 * larger than requested because of alignment.
 *
 * If allocation is nonzero, objects returned from tango2_pool_alloc() won't
 * cross that size boundary.  This is useful for devices which have
 * addressing restrictions on individual DMA transfers, such as not crossing
 * boundaries of 4KBytes.
 */
struct tango2_pool *
tango2_pool_create (const char *name, /*struct pci_dev *pdev,*/
	size_t size, size_t align, size_t allocation, int flags)
{
	struct tango2_pool		*retval;


	//printk("tango pool create, %s\n", name);

	if (align == 0)
		align = 1;
	if (size == 0)
		return 0;
	else if (size < align)
		size = align;
	else if ((size % align) != 0) {
		size += align + 1;
		size &= ~(align - 1);
	}

	if (allocation == 0) {
		if (PAGE_SIZE < size)
			allocation = size;
		else
			allocation = PAGE_SIZE;
		// FIXME: round up for less fragmentation
	} else if (allocation < size)
		return 0;

	if (!(retval = kmalloc (sizeof *retval, flags)))
		return retval;

	memset(retval, 0, sizeof *retval);

#ifdef	CONFIG_TANGO2POOL_DEBUG
	flags |= SLAB_POISON;
#endif

	strncpy (retval->name, name, sizeof retval->name);
	retval->name [sizeof retval->name - 1] = 0;

	//retval->dev = pdev;
	INIT_LIST_HEAD (&retval->page_list);
	spin_lock_init (&retval->lock);
	retval->size = size;
	retval->flags = flags;
	retval->allocation = allocation;
	retval->blocks_per_page = allocation / size;
	init_waitqueue_head (&retval->waitq);

#ifdef CONFIG_TANGO2POOL_DEBUG
	printk (KERN_DEBUG "tango2 pool create %s/%s size %d, %d/page (%d alloc)\n",
		 NULL, retval->name, size,
		retval->blocks_per_page, allocation);
#endif

	return retval;
}


static struct tango2_page * tango2_pool_alloc_page (struct tango2_pool *pool, int mem_flags)
{
	struct tango2_page	*page;
	int		mapsize;

	//printk("tango2_pool_alloc_page\n");
	mapsize = pool->blocks_per_page;
	mapsize = (mapsize + BITS_PER_LONG - 1) / BITS_PER_LONG;
	mapsize *= sizeof (long);

	page = (struct tango2_page *) kmalloc (mapsize + sizeof *page, mem_flags);
	if (!page)
		return 0;

	memset(page, 0, (mapsize + sizeof *page));

	page->vaddr = tango2_alloc_consistent (/*pool->dev,*/
					    pool->allocation,
					    &page->dma);
	if (page->vaddr) {
		memset (page->bitmap, 0xff, mapsize);	// bit set == free
		if (pool->flags & SLAB_POISON)
			memset (page->vaddr, POOL_POISON_BYTE, pool->allocation);
		list_add (&page->page_list, &pool->page_list);
	} else {
		kfree (page);
		page = 0;
	}
	return page;
}



static void
pool_free_page (struct tango2_pool *pool, struct tango2_page *page)
{
	dma_addr_t	dma = page->dma;

	if (pool->flags & SLAB_POISON)
		memset (page->vaddr, POOL_POISON_BYTE, pool->allocation);
	tango2_free_consistent (/*pool->dev,*/ pool->allocation, page->vaddr, dma);
	list_del (&page->page_list);
	kfree (page);
}


static inline int
tango2_is_page_busy (int blocks, unsigned long *bitmap)
{
        while (blocks > 0) {
                if (*bitmap++ != ~0UL)
                        return 1;
                blocks -= BITS_PER_LONG;
        }
        return 0;
}


/**
 * tango2_pool_destroy - destroys a pool of memory blocks.
 * @pool: pool that will be destroyed
 *
 * Caller guarantees that no more memory from the pool is in use,
 * and that nothing will try to use the pool after this call.
 */
void
tango2_pool_destroy (struct tango2_pool *pool)
{
	unsigned long		flags;

#ifdef CONFIG_TANGO2POOL_DEBUG
	printk (KERN_DEBUG "pool destroy %s/%s\n",
		/*pool->dev ? pool->dev->slot_name : */NULL,
		pool->name);
#endif

	spin_lock_irqsave (&pool->lock, flags);
	while (!list_empty (&pool->page_list)) {
		struct tango2_page		*page;
		page = list_entry (pool->page_list.next,
				struct tango2_page, page_list);
		if (tango2_is_page_busy (pool->blocks_per_page, page->bitmap)) {
			printk (KERN_ERR "tango2_pool_destroy %s/%s, %p busy\n",
				NULL,
				pool->name, page->vaddr);
			/* leak the still-in-use consistent memory */
			list_del (&page->page_list);
			kfree (page);
		} else
			pool_free_page (pool, page);
	}
	spin_unlock_irqrestore (&pool->lock, flags);
	kfree (pool);
}


/**
 * tango2_pool_alloc - get a block of consistent memory
 * @pool: pool that will produce the block
 * @mem_flags: SLAB_KERNEL or SLAB_ATOMIC
 * @handle: pointer to dma address of block
 *
 * This returns the kernel virtual address of a currently unused block,
 * and reports its dma address through the handle.
 * If such a memory block can't be allocated, null is returned.
 */
void *
tango2_pool_alloc (struct tango2_pool *pool, int mem_flags, dma_addr_t *handle)
{
	unsigned long		flags;
	struct list_head	*entry;
	struct tango2_page		*page;
	int			map, block;
	size_t			offset;
	void			*retval;

//	printk("tango2_pool_alloc mem_flags=0x%x", mem_flags);
restart:
	spin_lock_irqsave (&pool->lock, flags);
	list_for_each (entry, &pool->page_list) {
		int		i;
		page = list_entry (entry, struct tango2_page, page_list);
		/* only cachable accesses here ... */
		for (map = 0, i = 0;
				i < pool->blocks_per_page;
				i += BITS_PER_LONG, map++) {
			if (page->bitmap [map] == 0)
				continue;
			block = ffz (~ page->bitmap [map]);
			if ((i + block) < pool->blocks_per_page) {
				clear_bit (block, &page->bitmap [map]);
				offset = (BITS_PER_LONG * map) + block;
				offset *= pool->size;
				goto ready;
			}
		}
	}
	if (!(page = tango2_pool_alloc_page (pool, mem_flags))) {
		if (mem_flags == SLAB_KERNEL) {
			DECLARE_WAITQUEUE (wait, current);

			current->state = TASK_INTERRUPTIBLE;
			add_wait_queue (&pool->waitq, &wait);
			spin_unlock_irqrestore (&pool->lock, flags);

			schedule_timeout (POOL_TIMEOUT_JIFFIES);

			current->state = TASK_RUNNING;
			remove_wait_queue (&pool->waitq, &wait);
			goto restart;
		}
		retval = 0;
		goto done;
	}

	clear_bit (0, &page->bitmap [0]);
	offset = 0;
ready:
	retval = offset + page->vaddr;
	*handle = offset + page->dma;
done:
	spin_unlock_irqrestore (&pool->lock, flags);


//	printk("=0x%x dma_handle=0x%x offset=0x%x page->dma=0x%x\n", retval, *handle, offset, page->dma);
	return retval;
}


static struct tango2_page *
pool_find_page (struct tango2_pool *pool, dma_addr_t dma)
{
	unsigned long		flags;
	struct list_head	*entry;
	struct tango2_page		*page;

	spin_lock_irqsave (&pool->lock, flags);
	list_for_each (entry, &pool->page_list) {
		page = list_entry (entry, struct tango2_page, page_list);
		if (dma < page->dma)
			continue;
		if (dma < (page->dma + pool->allocation))
			goto done;
	}
	page = 0;
done:
	spin_unlock_irqrestore (&pool->lock, flags);
	return page;
}


/**
 * tango2_pool_free - put block back into  pool
 * @pool: the  pool holding the block
 * @vaddr: virtual address of block
 * @dma: dma address of block
 *
 * Caller promises neither device nor driver will again touch this block
 * unless it is first re-allocated.
 */
void
tango2_pool_free (struct tango2_pool *pool, void *vaddr, dma_addr_t dma)
{
	struct tango2_page		*page;
	unsigned long		flags;
	int			map, block;

	if ((page = pool_find_page (pool, dma)) == 0) {
		printk (KERN_ERR "tango2_pool_free %s/%s, %p/%x (bad dma)\n",
			NULL,
			pool->name, vaddr, (int) (dma & 0xffffffff));
		return;
	}
#ifdef	CONFIG_TANGO2POOL_DEBUG
	if (((dma - page->dma) + (void *)page->vaddr) != vaddr) {
		printk (KERN_ERR "tango2_pool_free %s/%s, %p (bad vaddr)/%x\n",
			NULL,
			pool->name, vaddr, (int) (dma & 0xffffffff));
		return;
	}
#endif

	block = dma - page->dma;
	block /= pool->size;
	map = block / BITS_PER_LONG;
	block %= BITS_PER_LONG;

#ifdef	CONFIG_TANGO2POOL_DEBUG
	if (page->bitmap [map] & (1UL << block)) {
		printk (KERN_ERR "tango2_pool_free %s/%s, dma %x already free\n",
			NULL,
			pool->name, dma);
		return;
	}
#endif
	if (pool->flags & SLAB_POISON)
		memset (vaddr, POOL_POISON_BYTE, pool->size);

	spin_lock_irqsave (&pool->lock, flags);
	set_bit (block, &page->bitmap [map]);
	if (waitqueue_active (&pool->waitq))
		wake_up (&pool->waitq);
	/*
	 * Resist a temptation to do
	 *    if (!is_page_busy(bpp, page->bitmap)) pool_free_page(pool, page);
	 * it is not interrupt safe. Better have empty pages hang around.
	 */
	spin_unlock_irqrestore (&pool->lock, flags);
}



void tango2_usb_init(void)
{
        int  i;
        unsigned long temp;

        /*check see if it's inited*/
        temp = gbus_read_uint32(pGBus, TANGO2_USB_CTL_STATUS_REG_BASE + 0x0);
        if(temp && (1<<19)){
                printk("Tango2 USB was initialized.\n");
                return;
        }
        else
                printk("Tango2 USB initializing...\n");

        /*use internal clock*/
        //gbus_write_uint32(pGBus,  TANGO2_USB_CTL_STATUS_REG_BASE + 0xc, 0xf991a);
        /* external clock at 24MHZ*/
        //gbus_write_uint32(pGBus , TANGO2_USB_CTL_STATUS_REG_BASE + 0xc, 0xf9931);
        /* external clock at 48Mhz*/
        // gbus_write_uint32(pGBus, TANGO2_USB_CTL_STATUS_REG_BASE + 0xc, 0xf9932);

        //udelay (1000);

        /*reseting phy*/
        temp = gbus_read_uint32(pGBus, TANGO2_USB_CTL_STATUS_REG_BASE + 0x0);
        DBG("Performing PHY Reseting 0x%x...\n", temp);
        gbus_write_uint32(pGBus, TANGO2_USB_CTL_STATUS_REG_BASE + 0x0, temp | 0x01);
        udelay (30);
        gbus_write_uint32(pGBus, TANGO2_USB_CTL_STATUS_REG_BASE + 0x0, temp);

        /*reseting ohci*/
        wait_ms(5);
        temp = gbus_read_uint32(pGBus, TANGO2_OHCI_BASE_ADDR + 0x08);
        DBG("Performing USB OHCI Reseting 0x%x...\n", temp);
        gbus_write_uint32(pGBus, TANGO2_OHCI_BASE_ADDR + 0x08,  temp | 0x01);
        wait_ms(5);
        //gbus_write_uint32(pGBus, TANGO2_OHCI_BASE_ADDR + 0x08, temp);
        //temp = gbus_read_uint32(pGBus, TANGO2_OHCI_BASE_ADDR + 0x08);

        /*reseting ohci dpll, it says the bit is for simulation*/
        temp = gbus_read_uint32(pGBus, TANGO2_USB_CTL_STATUS_REG_BASE + 0x0);
        DBG("Performing USB OHCI DPLL Reseting 0x%x...\n", temp);
        gbus_write_uint32(pGBus,  TANGO2_USB_CTL_STATUS_REG_BASE + 0x0, temp | (1<<19));
        wait_ms(1);

        for(i = 0; i < 4; i++){
                temp = gbus_read_uint32(pGBus, TANGO2_USB_CTL_STATUS_REG_BASE + i*4);
                DBG("Tango2 USB register %d = 0x%x\n", i, temp);
        }

        return;
}




/* Pool allocator */
EXPORT_SYMBOL (tango2_pool_create);
EXPORT_SYMBOL (tango2_pool_destroy);
EXPORT_SYMBOL (tango2_pool_alloc);
EXPORT_SYMBOL (tango2_pool_free);
EXPORT_SYMBOL(tango2_alloc_consistent);
EXPORT_SYMBOL(tango2_free_consistent);
EXPORT_SYMBOL(tango2_map_single);
EXPORT_SYMBOL(tango2_unmap_single);
EXPORT_SYMBOL(tango2_usb_init);
